(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: value_type.ml,v 1.10 2014/01/10 19:22:10 fclement Exp $ *)

open Value_type_def;;

let bool_couple_type = couple_type (bool_type, bool_type)
and bool_triple_type = triple_type (bool_type, bool_type, bool_type)
and int_couple_type = couple_type (int_type, int_type)
and int_triple_type = triple_type (int_type, int_type, int_type)
and native_int_couple_type =
  couple_type (native_int_type, native_int_type)
and native_int_triple_type =
  triple_type (native_int_type, native_int_type, native_int_type)
and int32_couple_type = couple_type (int32_type, int32_type)
and int32_triple_type = triple_type (int32_type, int32_type, int32_type)
and int64_couple_type = couple_type (int64_type, int64_type)
and int64_triple_type = triple_type (int64_type, int64_type, int64_type)
and float_couple_type = couple_type (float_type, float_type)
and float_couple_bin_type = couple_type (float_bin_type, float_bin_type)
and float_triple_type = triple_type (float_type, float_type, float_type)
and float_triple_bin_type =
    triple_type (float_bin_type, float_bin_type, float_bin_type)
;;

let rec scalar_type_to_ob ob = function
  | Bool_type -> Buffer.add_string ob "%B"
  | String_type -> Buffer.add_string ob "%S"
  | Int_type -> Buffer.add_string ob "%i"
  | Native_int_type -> Buffer.add_string ob "%ni"
  | Int32_type -> Buffer.add_string ob "%li"
  | Int64_type -> Buffer.add_string ob "%Li"
  | Float_type ->
      Buffer.add_string ob
        (if Options_pio.get_binary_mode () then "%bf" else "%f")
  | Float_bin_type -> Buffer.add_string ob "%bf"
  | Couple_type (t1, t2) ->
      Buffer.add_char ob '(';
      scalar_type_to_ob ob t1;
      Buffer.add_string ob ", ";
      scalar_type_to_ob ob t2;
      Buffer.add_char ob ')';
  | Triple_type (t1, t2, t3) ->
      Buffer.add_char ob '(';
      scalar_type_to_ob ob t1;
      Buffer.add_string ob ", ";
      scalar_type_to_ob ob t2;
      Buffer.add_string ob ", ";
      scalar_type_to_ob ob t3;
      Buffer.add_char ob ')';
;;

let value_type_to_ob ob = function
  | Scalar_type sty ->
      scalar_type_to_ob ob sty
  | Vector_type sty ->
      Buffer.add_string ob "[1";
      scalar_type_to_ob ob sty;
      Buffer.add_string ob "1]";
  | Matrix_type sty ->
      Buffer.add_string ob "[2";
      scalar_type_to_ob ob sty;
      Buffer.add_string ob "2]";
;;

let scalar_type_of_ib ib =
  let rec sos ib =
    Scanf.bscanf ib "%c" (function
    | '%' ->
      Scanf.bscanf ib "%c" (function
      | 'B' -> bool_type
      | 'S' -> string_type
      | 'i' -> int_type
      | 'f' -> float_type
      | 'b' -> Scanf.bscanf ib "%[f]" (function
        | "f" -> float_bin_type
        | s ->
          failwith
            (Printf.sprintf
               "scalar_type_of_ib: unknown binary format %%b%s"
               s))
      | c ->
        failwith
          (Printf.sprintf "scalar_type_of_ib: unknown format %%%c" c))
    | '(' (* ) *) ->
      let rec loop ts =
        Scanf.bscanf ib "%c" (function
        | (* ( *) ')' ->
          begin match ts with
          | [t1] -> t1
          | [t1; t2] -> couple_type (t2, t1)
          | [t1; t2; t3] -> triple_type (t3, t2, t1)
          | _ ->
            failwith
              "scalar_type_of_ib: tuple are restricted to couple or triple" end
        | ',' ->
          Scanf.bscanf ib " " ();
          let t2 = sos ib in
          loop (t2 :: ts)
        | c ->
          failwith
            (Printf.sprintf
               "scalar_type_of_ib: ',' or ')' expected; found %c" c)) in
      let t1 = sos ib in
      loop [t1]
    | c ->
      failwith
        (Printf.sprintf
           "scalar_type_of_ib: '%%' or '(' expected; found %c" c)) in
  sos ib
;;

let value_type_of_ib ib =
  Scanf.bscanf ib "%0c" (function
  | '[' (* ] *) -> Scanf.bscanf ib "%_c%d" (function
    | 1 ->
      let tv = vector_type (scalar_type_of_ib ib) in
      Scanf.bscanf ib "1]" ();
      tv
    | 2 ->
      let tm = matrix_type (scalar_type_of_ib ib) in
      Scanf.bscanf ib "2]" ();
      tm
    | p ->
      failwith
        (Printf.sprintf
           "value_type_of_ib: multi-dimensional array of \
            dimension %i not yet implemented"
           p))
  | _ -> scalar_type (scalar_type_of_ib ib));;

let to_ob = value_type_to_ob;;
let of_ib = value_type_of_ib;;
